(in-package :araneida)

;;; Make emacs evaluate
;;;  (put 'with-ignored-signals 'lisp-indent-function 'defun)
;;; for sane indenting in HANDLE-a-REQUEST

(defun accept-handler (listener)
  (let ((socket (sockets:socket-accept listener)))
    (handle-a-request
     (sockets:socket-make-stream socket :element-type 'base-char
				 :input t :output t :buffering :full)
     socket)))

(defun install-serve-event-handlers ()
  (let* ((ports (remove-duplicates
                 (mapcar #'server-port *exported-servers*)))
         (listeners
          (mapcar (lambda (x)
                    (listen-to-inet-port :port x :reuse 1))
                  ports)))
    (unless listeners (error "No sockets to listen on"))
    (dolist (socket listeners)
      (event:add-fd-handler (sockets:socket-file-descriptor  socket)
			    :input (lambda (fd) (accept-handler socket))))))
				     
(eval-when (:load-toplevel :execute)
  (sb-sys:ignore-interrupt :sigpipe))

(defun handle-a-request (stream socket)
  (let* ((request-line (read-folded-line stream nil))
         (headers (read-headers stream))
         (content-length (parse-integer
                          (or (header-value :content-length headers) "0")))
	 (body (and (> content-length 0)
		    (make-array content-length :element-type
				(stream-element-type stream))))
         (len (and (> content-length 0)
		   (read-sequence body stream))))
    (destructuring-bind
	  (method url-string protocol) (split request-line 3 '(#\Space))
      (declare (ignore protocol))
      (let* ((parsed-body (if body (parse-body body '(#\&) len) nil))
             (url (merge-url
                   (server-base-url (server-for-name-colon-port
                                     (header-value :host headers)))
                   url-string))
             (r (make-instance
                 'request :url url
		 :urlstring (urlstring url)
                 :method (intern (string-upcase method) 'keyword)
                 :body parsed-body
                 :stream stream :headers headers :socket socket)))
	(handle-request *root-handler* r)
	;; when we support persistent connectionms, we'll have to remove
	;; this close operation
	(let* ((s (request-stream r)))
	  (forcibly-close-stream s))))))

(defun parse-body (body-string &optional (delimiters '(#\&)) end)
  "Parse BODY-STRING returning list of (var val) pairs"
  (mapcar (lambda (x)
            (mapcar #'urlstring-unescape (split-sequence #\= x :count 2)))
          (split-sequence-if (lambda (x) (member x delimiters))
			     body-string :end end)))

(defun read-folded-line (stream &optional eof-error-p eof-value)
  "Read a complete logical header line, including folded continuation lines."
  (with-output-to-string (o)
    (loop
     (let* ((l (read-line stream eof-error-p eof-value))
	    (end (position-if (lambda (x) (or (eql x (code-char 10))
					      (eql x (code-char 13))))
			      l))
	    (next (and (if end (> end 0) t)
		       (> (length l) 0)
		       (peek-char nil stream nil nil))))
       (write-sequence l o :end end)
       (unless (or (eql next #\Space) (eql next #\Tab)) (return))))))

;;; Unless you're me (and probably not even then), you should be
;;; using REQUEST-HEADER not this:

(defun header-value (name header-list)
  "Get the value of the header named NAME in HEADER-LIST"
  (cadr (assoc name header-list)))

(defun read-headers (stream)
  (let ((headers nil))
    (do ((line (read-folded-line stream t) (read-folded-line stream t)))
        ((or (not line) (string= line "")) headers)
      (destructuring-bind (keyword value) (split line 2 '(#\:))
        ;; RFC 1945: "Each header field consists of a name followed
        ;; immediately by a colon (":"), a single space (SP)
        ;; character, and the field value. Field names are case-
        ;; insensitive." - so we need to lose the first char of value
        (let* ((value (subseq value 1 nil))
               (keyword-symbol (intern (string-upcase keyword) :keyword))
               (keyword-value (assoc keyword-symbol  headers)))
          (if keyword-value
              (setf (cdr keyword-value)
                    (append (cdr keyword-value) (list value)))
            (setf headers (acons keyword-symbol (list value) headers))))))
    headers))
        
